#pragma once

#include <iostream>

// Class for representing ranges of characters, usually for the
// purpose of comparing them. The idea is that we have constructors for
// a bunch of different ways that strings might be specified, and each
// gets converted to a string_view [begin, end).
//
// Enforces invariant that begin <= end; throws otherwise.
class string_view
{
public:
    // Constructs an empty string_view.
    string_view();

    // Constructs a string_view from `begin` and `end` directly.
    string_view(const char* begin, const char* end);

    // Constructs a string_view from the start and the size.
    string_view(const char*, size_t);

    // Constructs a string_view from a '\0'-terminated C-style string.
    template <class T>
    string_view(T const* const&);

    // Constructs a string_view from a string literal using its static size.
    template <class T, size_t N>
    string_view(T const (&s) [N]);

    /*
     * Getters
     */

    // Returns the number of characters.
    size_t size() const;

    // Returns whether this string_view is empty.
    bool empty() const;

    // Gets the pointer to the first character.
    const char* begin() const;

    // Gets the pointer one past the last character.
    const char* end() const;

    // Indexes.
    //
    // PRECONDITION (unchecked): index < size()
    char operator[](size_t index) const;

    // Returns the substring from index `start` to the end.
    //
    // PRECONDITION (asserted):
    //  - start is in bounds
    string_view substr(size_t start) const;

    // Returns the substring from index `start` of length `size`.
    //
    // PRECONDITION (asserted):
    //  - [start, start + size) is in bounds
    string_view substr(size_t start, size_t size) const;

private:
    /*
     * Member variables
     */

    const char* begin_;
    const char* end_;
    // INVARIANT: begin_ <= end_
};

// Overloads == for `string_view`s.
bool operator==(string_view, string_view);

// Overloads != for `string_view`s.
bool operator!=(string_view, string_view);

template <class T>
string_view::string_view(T const* const& s)
        : string_view(s, std::strlen(s))
{ }

// Templates need to be defined in the .h file, not the .cpp file.
// (We subtract 1 from N because N will include the string literal's
// '\0' terminator.)
template <class T, size_t N>
string_view::string_view(T const (&s) [N])
        : string_view(s, N - 1)
{ }

// Overloads stream insertion (printing)
std::ostream& operator<<(std::ostream&, string_view);

